#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>

// This kernel computes each output element for transposed convolution
// and manually unrolls loops to reduce loop overhead. It assumes the kernel dimensions
// (kernel_h, kernel_w) are small so that unrolling them is beneficial, and it unrolls the inner
// loop over input channels in groups of 4 when possible.

template <typename scalar_t>
__global__ void conv_transpose2d_gather_fused_unroll_kernel(
    const scalar_t* __restrict__ input,
    const scalar_t* __restrict__ weight,
    const scalar_t* __restrict__ bias, // may be nullptr
    scalar_t* __restrict__ output,
    const int batch_size,
    const int in_channels,
    const int in_height,
    const int in_width,
    const int out_channels,
    const int kernel_h,
    const int kernel_w,
    const int stride,
    const int padding,
    const int output_padding, // used only for output size computation
    const int groups,
    const int dilation,
    const int out_height,
    const int out_width
) {
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    int total_elements = batch_size * out_channels * out_height * out_width;
    if (idx >= total_elements) return;

    // Unravel the flat index into 4D indices: (b, oc, oh, ow)
    int tmp = idx;
    int ow = tmp % out_width;
    tmp /= out_width;
    int oh = tmp % out_height;
    tmp /= out_height;
    int oc = tmp % out_channels;
    int b = tmp / out_channels;

    // Determine group related indices
    int out_channels_per_group = out_channels / groups;
    int g = oc / out_channels_per_group;
    int oc_group = oc % out_channels_per_group;
    int in_channels_per_group = in_channels / groups;
    int in_channel_start = g * in_channels_per_group;

    // Initialize the accumulation with bias if provided
    scalar_t sum = (bias != nullptr) ? bias[oc] : static_cast<scalar_t>(0);

    // Pre-calculate strides
    int in_image_size = in_height * in_width;

    // Loop over kernel height and width with unrolling
    #pragma unroll
    for (int kh = 0; kh < kernel_h; ++kh) {
        int h_offset = oh + padding - kh * dilation;
        if (h_offset < 0 || h_offset % stride != 0) continue;
        int h_in = h_offset / stride;
        if (h_in < 0 || h_in >= in_height) continue;

        #pragma unroll
        for (int kw = 0; kw < kernel_w; ++kw) {
            int w_offset = ow + padding - kw * dilation;
            if (w_offset < 0 || w_offset % stride != 0) continue;
            int w_in = w_offset / stride;
            if (w_in < 0 || w_in >= in_width) continue;

            int input_base = b * (in_channels * in_image_size);
            // Manual unrolling for the inner loop over input channels in groups of 4
            int unroll_end = (in_channels_per_group / 4) * 4;
            for (int ic = 0; ic < unroll_end; ic += 4) {
                int cur_ic0 = in_channel_start + ic;
                int cur_ic1 = cur_ic0 + 1;
                int cur_ic2 = cur_ic0 + 2;
                int cur_ic3 = cur_ic0 + 3;

                int input_idx0 = input_base + cur_ic0 * in_image_size + h_in * in_width + w_in;
                int input_idx1 = input_base + cur_ic1 * in_image_size + h_in * in_width + w_in;
                int input_idx2 = input_base + cur_ic2 * in_image_size + h_in * in_width + w_in;
                int input_idx3 = input_base + cur_ic3 * in_image_size + h_in * in_width + w_in;

                // Weight indexing: [in_channels, out_channels_per_group, kernel_h, kernel_w]
                int weight_idx0 = cur_ic0 * (out_channels_per_group * kernel_h * kernel_w) +
                                  oc_group * (kernel_h * kernel_w) + kh * kernel_w + kw;
                int weight_idx1 = cur_ic1 * (out_channels_per_group * kernel_h * kernel_w) +
                                  oc_group * (kernel_h * kernel_w) + kh * kernel_w + kw;
                int weight_idx2 = cur_ic2 * (out_channels_per_group * kernel_h * kernel_w) +
                                  oc_group * (kernel_h * kernel_w) + kh * kernel_w + kw;
                int weight_idx3 = cur_ic3 * (out_channels_per_group * kernel_h * kernel_w) +
                                  oc_group * (kernel_h * kernel_w) + kh * kernel_w + kw;

                sum += input[input_idx0] * weight[weight_idx0] +
                       input[input_idx1] * weight[weight_idx1] +
                       input[input_idx2] * weight[weight_idx2] +
                       input[input_idx3] * weight[weight_idx3];
            }
            // Process remaining channels if any
            for (int ic = unroll_end; ic < in_channels_per_group; ++ic) {
                int cur_ic = in_channel_start + ic;
                int input_idx = input_base + cur_ic * in_image_size + h_in * in_width + w_in;
                int weight_idx = cur_ic * (out_channels_per_group * kernel_h * kernel_w) +
                                 oc_group * (kernel_h * kernel_w) + kh * kernel_w + kw;
                sum += input[input_idx] * weight[weight_idx];
            }
        }
    }

    int output_idx = b * (out_channels * out_height * out_width) +
                     oc * (out_height * out_width) +
                     oh * out_width + ow;
    output[output_idx] = sum;
}


// Forward function: sets up output tensor and launches the unrolled kernel

torch::Tensor forward(
    torch::Tensor input,
    torch::Tensor weight,
    torch::optional<torch::Tensor> bias,
    int stride,
    int padding,
    int output_padding,
    int groups,
    int dilation = 1
) {
    TORCH_CHECK(input.is_cuda(), "input must be a CUDA tensor");
    TORCH_CHECK(weight.is_cuda(), "weight must be a CUDA tensor");
    TORCH_CHECK(input.dim() == 4, "Input must be 4D");
    TORCH_CHECK(weight.dim() == 4, "Weight must be 4D");

    const int batch_size = input.size(0);
    const int in_channels = input.size(1);
    const int in_height = input.size(2);
    const int in_width = input.size(3);
    
    // Weight shape assumed: [in_channels, out_channels/groups, kernel_h, kernel_w]
    const int out_channels = weight.size(1) * groups; // weight.size(1) is out_channels per group
    const int kernel_h = weight.size(2);
    const int kernel_w = weight.size(3);

    // Calculate output dimensions using transposed convolution formula
    const int out_height = (in_height - 1) * stride - 2 * padding + dilation * (kernel_h - 1) + output_padding + 1;
    const int out_width  = (in_width - 1) * stride - 2 * padding + dilation * (kernel_w - 1) + output_padding + 1;

    auto output = torch::zeros({batch_size, out_channels, out_height, out_width}, input.options());

    const int total_elements = batch_size * out_channels * out_height * out_width;
    const int threads = 256;
    const int blocks = (total_elements + threads - 1) / threads;

    AT_DISPATCH_FLOATING_TYPES(input.scalar_type(), "conv_transpose2d_gather_fused_unroll_cuda", ([&] {
        const scalar_t* bias_data = (bias.has_value() && bias->defined()) ? bias->data_ptr<scalar_t>() : nullptr;
        conv_transpose2d_gather_fused_unroll_kernel<scalar_t><<<blocks, threads>>>(
            input.data_ptr<scalar_t>(),
            weight.data_ptr<scalar_t>(),
            bias_data,
            output.data_ptr<scalar_t>(),
            batch_size,
            in_channels,
            in_height,
            in_width,
            out_channels,
            kernel_h,
            kernel_w,
            stride,
            padding,
            output_padding,
            groups,
            dilation,
            out_height,
            out_width
        );
    }));

    return output;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward, "Fused Transposed 2D Convolution with Loop Unrolling (CUDA)",
          py::arg("input"),
          py::arg("weight"),
          py::arg("bias") = py::none(),
          py::arg("stride"),
          py::arg("padding"),
          py::arg("output_padding"),
          py::arg("groups"),
          py::arg("dilation") = 1);
}
